import { Component, writeNotify } from "pi_ucore/util/ecs";

export class DataC extends Component {
    area: number; // 区域编号
    layer: number; // 层级
    pos: [number, number]; // 坐标(左下角) [x,y]

    constructor(area: number, layer: number, pos: [number, number]) {
        super()
        this.area = area;
        this.layer = layer;
        this.pos = pos;
    }
}

export class SizeC extends Component {
    width: number = 1; // 宽
    height: number = 1; // 高

    constructor(size: [number, number]) {
        super();
        this.width = size[0];
        this.height = size[1];
    }
}

export class StateC extends Component {
    _value: DataState = DataState.Normal; // 物品状态
    oldValue: DataState; // 上次的状态

    @writeNotify
    set value(state: DataState) {
        this.oldValue = this._value
        this._value = state;
    }

    get value() {
        return this._value;
    }

    // 正在被影响
    public isExploding() {
        return this.value === DataState.WaitExplode || this.value === DataState.AroundExplode;
    }
}

// 层级类型
export enum LayerType {
    Lawn = 0, // 草坪
    Data = 1, // 物品
    Lock = 2, // 锁 阻止掉落
}

// 物品状态
export enum DataState {
    Normal = 1, // 正常
    WaitMove = 2, // 等待移动
    Moving = 3, // 移动中
    WaitExplode = 4, // 等待爆炸
    AroundExplode = 5, // 周围爆炸影响
    Exploded = 6, // 爆炸完
}

// 方向
export enum DirectionType {
    Up = 1, // 上
    Down = 2, // 下
    Left = 3, // 左
    Right = 4, // 右
    SlowDown = 5, // 缓慢向下
    Other = 6, // 其他 坐标已处理
}
