import { writeNotify } from "pi_ucore/util/ecs";
import { MatchMapCfg } from "../xls/map/map.struct";
import { MatchMapACfg } from "../xls/map/map_a.struct";
import { MatchMapBCfg } from "../xls/map/map_b.struct";

export class GlobalState {
    areaIndexs: Map<number, number[][][]>; // 每个区域的索引数据 [区域，entity数组(layer,x,y)]
    areaMap: Map<number, [number, number]>; // 区域地图 [区域，[宽，高]]
    curArea: number; // 当前操作区域
    mapCfg: MatchMapCfg | MatchMapACfg | MatchMapBCfg; // 地图配置
    seed: number; // 随机种子
    eventDataList: [number, string][] = []; // 事件物品数组 [物品类型，奖品ID] 初始随机，有固定数量，用完即止
    @writeNotify
    gameState: GameState = GameState.Normal; // 游戏状态
    cannotBombList: number[] = []; // 不能被炸的物品列表
    waitSecondPart: boolean = false; // 等待第二段爆炸

    constructor() {
        this.areaIndexs = new Map();
        this.areaMap = new Map();
        this.curArea = 1;
    }

    public canOperate() {
        return this.gameState === GameState.DropEnd || this.gameState === GameState.Normal;
    }
}

export enum GameState {
    Normal = 0, // 正常
    PlayAnim = 1, // 播动画
    AnimEnd = 2, // 动画结束
    Droping = 3, // 掉落
    DropEnd = 4, // 掉落结束
    Waiting = 5, // 等待
}