import { init } from "pi_scene/scene/util/init_util";
import { initPosition } from "pi_scene/scene/util/common_util";
import { Entity, World } from "pi_ucore/util/ecs";
import { DataC, DataState, LayerType, SizeC, StateC } from "../component/data";
import { DropSys } from "../system/drop_sys";
import { TimeStepSys } from "../system/time_step_sys";
import { ExplodeSys } from "../system/explode_sys";
import { AnimSys } from "../system/anim_sys";
import { frameRun } from "pi_scene/scene/util/init_util";
import { GameState, GlobalState } from "../single/global_state";
import { getCfg } from "pi_common_tools/util/db_tool";
import { MatchMapCfg } from "../xls/map/map.struct";
import { MatchMapACfg } from "../xls/map/map_a.struct";
import { MatchMapBCfg } from "../xls/map/map_b.struct";
import { UINT32_MAX } from "pi_math/math";
import { DirectionC, MoveC, PositionC, SpeedC } from "pi_scene/scene/component/common";
import { AroundSys } from "../system/around_sys";
import { BombSys } from "../system/bomb_sys";
import { MatchSys } from "../system/match_sys";
import { ActionCmd } from "../single/action_cmd";
import { BombList, BombType } from "../single/bomb_list";
import { MedusaNeedBombTimes, MedusaPetrifactionInterval, MedusaPetrifactionCount, OverCountEventDataPro, OverCountEventData } from "../xls/base/base.struct";
import { FrameTime } from "pi_scene/scene/single/global";
import { ActionSys } from "../system/action_sys";
import { setThreeMathWorld, randomIndexByWeight, randomPos, getGrid } from "./base";
import { Calc } from "pi_common_tools/util/calc";
import { Rand } from "pi_math/rand";
import { AnimList } from "../single/anim_list";
import { ExpandSys } from "../system/expand_sys";
import { CollectSys } from "../system/collect_sys";
import { Collect } from "../single/collect";
import { TimeStep } from "../single/time_step";
import { GoalC, GoalType } from "../component/goal";
import { DataType, EventDataType, LockTypeC, NormalLockType, SpecialLockType, TypeC } from "../component/type";
import { ExplodeCount } from "../single/explode_count";
import { EventDataCfg } from "../xls/data_type/event.struct";
import { HeroSite } from "../single/hero_site";
import { DataTypeList } from "../single/data_type_list";
import { SynthesisDataCfg } from "../xls/data_type/synthesis.struct";
import { BombProgress } from "../single/bomb_progress";
import { TimeAwardDataCfg } from "../xls/data_type/time_award.struct";

// 初始化world autoMatch: 自动消除
export const initMatchWorld = (autoMatch: boolean = false, dispatcherName = "") => {
    const world = init(dispatcherName);
    initPosition(world, true, true, dispatcherName);
    world.registerComponent(Entity, DataC, 'data', true);
    world.registerComponent(Entity, TypeC, 'type', true);
    world.registerComponent(Entity, StateC, 'state', true);
    world.registerComponent(Entity, LockTypeC, 'lockType', true);
    world.registerComponent(Entity, SizeC, 'size');
    world.registerComponent(Entity, GoalC, 'goal');
    world.registerSingle(new GlobalState(), 'GlobalState');
    world.registerSingle(new ActionCmd(), 'ActionCmd');
    world.registerSingle(new BombList(), 'BombList');
    world.registerSingle(new Calc(), 'Calc');
    world.registerSingle(new AnimList(), 'AnimList');
    world.registerSingle(new Collect(), 'Collect');
    world.registerSingle(new TimeStep(), 'TimeStep');
    world.registerSingle(new ExplodeCount(), 'ExplodeCount');
    world.registerSingle(new HeroSite(), 'HeroSite');
    world.registerSingle(new DataTypeList(), 'DataTypeList');
    world.registerSingle(new BombProgress(), 'BombProgress');

    world.registerSystem('ExplodeSys', new ExplodeSys())
    world.registerSystem('DropSys', new DropSys())
    world.registerSystem('AnimSys', new AnimSys())
    world.registerSystem('BombSys', new BombSys())
    world.registerSystem('ExpandSys', new ExpandSys());
    world.registerSystem('ActionSys', new ActionSys())
    world.registerSystem('CollectSys', new CollectSys());
    world.registerSystem('AroundSys', new AroundSys())
    world.registerSystem('TimeStepSys', new TimeStepSys())
    let list = ["ExplodeSys", "AnimSys", "DropSys", "AroundSys", "TimeStepSys"];
    if (autoMatch) {
        world.registerSystem('MatchSys', new MatchSys())
        list.push("MatchSys")
    }

    // const frame = world.fetchSingle(FrameTime).data;
    // frame.frameInterval = 10;

    let d = world.getDispatcher(dispatcherName);
    d.build(list, world);
    setThreeMathWorld(world);

    return world;
}

// 初始化数据
export const initData = (world: World, mapId: number, heroList: number[], adTest?: string, failCount?: number[]) => {
    let cfg = !adTest ? getCfg(MatchMapCfg, mapId) : adTest === 'a' ? getCfg(MatchMapACfg, mapId) : getCfg(MatchMapBCfg, mapId);
    const globalState = world.fetchSingle(GlobalState).data;
    const bombList = world.fetchSingle(BombList).data;
    const calc = world.fetchSingle(Calc).data;
    const heroSite = world.fetchSingle(HeroSite).data;
    const dataTypeList = world.fetchSingle(DataTypeList).data;
    const collect = world.fetchSingle(Collect).data;

    if (failCount) collect.setFailCount(failCount);
    let seed = cfg.seed || window?.['testSeed'] || Math.ceil(Math.random() * UINT32_MAX);
    calc.rand = new Rand(seed);  // 用于其他所有随机
    dataTypeList.rand = new Rand(seed);  // 用于随机第一屏的所有物品 以及第二屏的基础物品
    let list = [];
    // 随机炸弹类型
    for (let i = 0; i < cfg.bombCount; i++) {
        list.push(BombType.FourBomb);
    }
    bombList.list = list;
    console.log('初始化炸弹列表', list, '随机种子', seed);

    globalState.mapCfg = cfg;
    globalState.seed = seed;
    let eventDataList = [];
    for (let [type, count, id] of cfg.eventData) {
        for (let i = 0; i < count; i++) {
            eventDataList.push([type, id]);
        }
    }
    for (let i = eventDataList.length; i < cfg.allCount; i++) {
        eventDataList.push(null)
    }
    dataTypeList.rand.shuffle(eventDataList);
    globalState.eventDataList = eventDataList;
    heroSite.heroList = heroList;

    let [w, h] = cfg.area1;
    let indexs = [];
    let lockLayer = initLockLayer(world, cfg, w, h);
    indexs[LayerType.Lock] = lockLayer;
    indexs[LayerType.Data] = initDataLayer(world, lockLayer, cfg, dataTypeList.rand, w, h);
    indexs[LayerType.Lawn] = initLawnLayer(world, lockLayer, cfg, dataTypeList.rand, w, h);
    replaceDataLayer(world, cfg, indexs[LayerType.Data]);

    globalState.areaMap.set(1, [w, h]);
    globalState.areaIndexs.set(1, indexs);
    console.log('initData============', mapId, indexs);
}

// 开始游戏
export const startMatch = (world: World) => {
    frameRun(world);
}

// 游戏结束
export const gameOver = (world: World) => {
    const frame = world.fetchSingle(FrameTime).data;
    frame.count = frame.maxCount;
}

// 创建物品
export const createOneData = (world: World, pos: [number, number], area: number, layer: number, size: [number, number] = [1, 1],
    needBombTimes: number = 1, type?: DataType, subType?: any, subTypeArg?: string, hideType?: any) => {

    const id = world.createEntity(Entity);
    const dataC = world.fetchComponent(Entity, DataC);
    dataC.insert(id, new DataC(area, layer, pos));

    initDataType(world, id, needBombTimes, type, subType, subTypeArg, hideType, pos);

    const goalC = world.fetchComponent(Entity, GoalC);
    // 美杜莎 随机石化格子
    if (type === DataType.Medusa) {
        goalC.insert(id, new GoalC(GoalType.RandomExpand, DataType.NormalLock, MedusaPetrifactionCount,
            { subType: NormalLockType.Petrifaction, times: MedusaPetrifactionInterval }));
    }
    // 信箱 产出信封
    if (subType === NormalLockType.MailBox) {
        goalC.insert(id, new GoalC(GoalType.OutPut, DataType.Mail, 999));
    }
    // 泡沫 往周围扩展泡沫
    if (subType === NormalLockType.Foam) {
        goalC.insert(id, new GoalC(GoalType.AroundExpand, DataType.NormalLock, 1, { subType: NormalLockType.Foam }))
    }
    // 合成物品 可合成高级物品
    if (type === DataType.SynthesisData) {
        let cfg = getCfg(SynthesisDataCfg, subType);
        goalC.insert(id, new GoalC(GoalType.Synthesis, cfg.sstype));
    }
    // 小丑鱼 产出鱼鳞
    if (subType === NormalLockType.Clownfish) {
        goalC.insert(id, new GoalC(GoalType.OutPut, DataType.FishScale, 999));
    }
    // 定时奖励物品 定时结束
    if (type === DataType.TimeAwardData) {
        const timeStep: TimeStep = world.fetchSingle(TimeStep).data;
        let cfg = getCfg(TimeAwardDataCfg, subType);
        const condition = { start: timeStep.step, end: timeStep.step + cfg.time };
        goalC.insert(id, new GoalC(GoalType.TimeAward, cfg.prop, cfg.count, condition));
    }
    const sizeC = world.fetchComponent(Entity, SizeC);
    sizeC.insert(id, new SizeC(size));
    const stateC = world.fetchComponent(Entity, StateC);
    stateC.insert(id, new StateC());
    const positionC = world.fetchComponent(Entity, PositionC);
    positionC.insert(id, new PositionC([pos[0], pos[1], 0]));
    const moveC = world.fetchComponent(Entity, MoveC);
    moveC.insert(id, new MoveC());
    const speedC = world.fetchComponent(Entity, SpeedC)
    speedC.insert(id, new SpeedC());
    const directionC = world.fetchComponent(Entity, DirectionC);
    directionC.insert(id, new DirectionC())

    return id;
}

// 删除物品
export const destroyOneData = (world: World, id: number) => {
    world.destroyEntity(Entity, id)
}

// ========================================本地方法=================================

// 初始化物品类型
const initDataType = (world: World, id: number, needBombTimes: number, type?: DataType, subType?: any, subTypeArg?: string, hideType?: any, pos: number[] = [0, 0]) => {
    const typeC = world.fetchComponent(Entity, TypeC);
    const globalState = world.fetchSingle(GlobalState).data;
    const calc = world.fetchSingle(Calc).data;
    const heroSite = world.fetchSingle(HeroSite).data;
    const dataTypeList = world.fetchSingle(DataTypeList).data;
    const collect = world.fetchSingle(Collect).data;
    const lockTypeC = world.fetchComponent(Entity, LockTypeC);
    const frameTime = world.fetchSingle(FrameTime).data;
    // 不指定类型 则随机基础物品
    const dataList = globalState.mapCfg.baseData.map(v => v[0]);
    const simpleLattice = collect.failCount.slice(1);
    const weightList = globalState.mapCfg.baseData.map(v => {
        if (!collect.failCount[0] || !globalState.mapCfg.downLimit[2] || frameTime.count <= 0) return v[1]
        if (simpleLattice.includes(v[0])) return v[1] + collect.failCount[0] * globalState.mapCfg.downLimit[2];
        return v[1]
    });
    if (!type && !hideType) {
        if (globalState.eventDataList.length) {
            let event = globalState.eventDataList.pop();
            if (event) {
                type = DataType.EventData;
                subType = event[0];
                subTypeArg = event[1];
            }
        } else {
            // 超出总数量按25%的概率补充剑盾
            let random = calc.rand.nextInt(0, 100);
            if (random < OverCountEventDataPro) {
                type = DataType.EventData;
                let eventDataList = JSON.parse(OverCountEventData);
                let weightList = eventDataList.map(v => v[2]);
                let event = randomIndexByWeight(eventDataList, weightList, calc.rand);
                subType = event[0];
                subTypeArg = event[1];
                // console.log('产出新的事件物品了', random, event);
            }
        }
    }
    if (!type) {
        // 基础物品替换类型
        let list = globalState.mapCfg.replaceDataType;
        let rand = dataTypeList.rand;
        if (list?.length > 0) {
            let ind = list.findIndex(v => v[0] === dataTypeList.list.length);
            if (ind > -1) {
                let [len, act, typ] = list[ind];
                if (act === 1) { // 增 本次不随机
                    type = typ;
                }
                else if (act === 2) { // 删 随机两次跳过第一次
                    randomIndexByWeight(dataList, weightList, rand);
                    type = randomIndexByWeight(dataList, weightList, rand);
                }
                else if (act === 3) { // 改 随机一次然后替换
                    randomIndexByWeight(dataList, weightList, rand);
                    type = typ;
                }
            }
        }
        // 随机基础物品
        if (!type) {
            type = randomIndexByWeight(dataList, weightList, rand);
        }
        dataTypeList.list.push([type, pos[0]]);
    }

    if (type === DataType.EventData && subType) {
        let cfg = getCfg(EventDataCfg, subType);
        needBombTimes = cfg?.needBombTimes || 1;
        // 炸开后替换为奖励物品
        if (cfg.atype) hideType = cfg.atype;
        // 英雄
        if (subType === EventDataType.Hero) {
            subTypeArg = randomHeroSite(heroSite, Number(subTypeArg)) as any
        }
    }
    // 不能炸的类型
    let types = [DataType.Cherry, DataType.Donut, DataType.SynthesisData, DataType.SpecialLock, DataType.AwardData];
    let cannotBomb = false;
    if (types.includes(type)) cannotBomb = true;
    typeC.insert(id, new TypeC(type, needBombTimes, cannotBomb, subType, subTypeArg, hideType));

    if (type !== DataType.NormalLock) return;
    // 阻挡横竖雷扩展
    let types1 = [NormalLockType.Biscuit];
    let blockBombExpansion = false;
    if (types1.includes(subType)) blockBombExpansion = true;
    // 周围炸影响自身
    let types2 = [NormalLockType.Jelly, NormalLockType.Container, NormalLockType.Biscuit, NormalLockType.MailBox, NormalLockType.Blinds, NormalLockType.Foam
        , NormalLockType.PearlTray, NormalLockType.Mermaid, NormalLockType.Ice];
    let aroundBombEffect = false;
    if (types2.includes(subType)) aroundBombEffect = true;
    // 不能压住物品
    let types3 = [NormalLockType.MailBox, NormalLockType.Container, NormalLockType.Biscuit, NormalLockType.Blinds, NormalLockType.PearlTray, NormalLockType.Mermaid, NormalLockType.Clownfish];
    let cannotHoldData = false;
    if (types3.includes(subType)) cannotHoldData = true;
    // 周围爆炸记录多次
    let types4 = [NormalLockType.MailBox, NormalLockType.PearlTray, NormalLockType.Mermaid];
    let recordMultipleTimes = false;
    if (types4.includes(subType)) recordMultipleTimes = true;
    lockTypeC.insert(id, new LockTypeC(blockBombExpansion, aroundBombEffect, cannotHoldData, recordMultipleTimes));
}

// 初始化锁层
const initLockLayer = (world: World, cfg: MatchMapCfg, w: number, h: number) => {
    let posList = [];
    let layerIds = [];
    for (let x = 0; x < w; x++) {
        layerIds[x] = []
        for (let y = 0; y < h; y++) {
            layerIds[x][y] = 0;
            if (cfg.fixPos.find(v => v[1] === x && v[2] === y)) continue;
            if (cfg.fixEventData.find(v => v[1] === x && v[2] === y)) continue;
            if (cfg.fixAwardData.find(v => v[1] === x && v[2] === y)) continue;
            posList.push([x, y]);
        }
    }
    let goalC = world.fetchComponent(Entity, GoalC);
    if (cfg.curtain.length > 0) { // 窗帘 全固定位置
        for (let i = 0; i < cfg.curtain.length; i++) {
            let { pos, size, collect } = cfg.curtain[i]
            let id = createOneData(world, pos, 1, LayerType.Lock, size, 0, DataType.SpecialLock, SpecialLockType.Curtain);
            goalC.insert(id, new GoalC(GoalType.Collect, collect[0], collect[1]));
            updatePosList(posList, layerIds, pos, size, id);
        }
    }
    if (cfg.blinds.length > 0) { // 百叶窗 全固定位置
        for (let i = 0; i < cfg.blinds.length; i++) {
            let { h, pos } = cfg.blinds[i]
            let id = createOneData(world, pos, 1, LayerType.Lock, [1, h], h, DataType.NormalLock, NormalLockType.Blinds);
            goalC.insert(id, new GoalC(GoalType.RollUp, null, h));
            updatePosList(posList, layerIds, pos, [1, h], id);
        }
    }
    if (cfg.bombDevice.length > 0) { // 炸弹装置 全固定位置
        for (let i = 0; i < cfg.bombDevice.length; i++) {
            let { pos, count, direction, list } = cfg.bombDevice[i];
            let id = createOneData(world, pos, 1, LayerType.Lock, [1, 1], 0, DataType.SpecialLock, SpecialLockType.BombDevice);
            goalC.insert(id, new GoalC(GoalType.SendBomb, DataType.SynthesisBomb, count, { direction, list: list.slice() }));
            updatePosList(posList, layerIds, pos, [1, 1], id);
        }
    }
    if (cfg.pearlTray.length > 0) { // 珍珠盘 全固定位置
        for (let i = 0; i < cfg.pearlTray.length; i++) {
            let { pos, size, count } = cfg.pearlTray[i];
            let id = createOneData(world, pos, 1, LayerType.Lock, size, count, DataType.NormalLock, NormalLockType.PearlTray);
            goalC.insert(id, new GoalC(GoalType.OutPut, DataType.Pearl, count));
            updatePosList(posList, layerIds, pos, size, id);
        }
    }
    if (cfg.mermaid.length > 0) { // 美人鱼 全固定位置
        for (let i = 0; i < cfg.mermaid.length; i++) {
            let { pos } = cfg.mermaid[i];
            let id = createOneData(world, pos, 1, LayerType.Lock, [2, 2], 999, DataType.NormalLock, NormalLockType.Mermaid);
            goalC.insert(id, new GoalC(GoalType.OutPut, DataType.MermaidLock, 8, null, 8));
            updatePosList(posList, layerIds, pos, [2, 2], id);
        }
    }

    let typeList = [];
    if (cfg.obstacle.length > 0) { // 障碍物
        for (let i = 1; i <= cfg.obstacle[0]; i++) {
            addTypeData(typeList, posList, DataType.SpecialLock, SpecialLockType.Obstacle, cfg.obstacle[i], 0)
        }
    }
    if (cfg.openSpaceObstacle.length > 0) { // 空地障碍物
        for (let i = 1; i <= cfg.openSpaceObstacle[0]; i++) {
            const subTypeArg = cfg.openSpaceObstacle[i] && cfg.openSpaceObstacle[i][2] ? `${cfg.openSpaceObstacle[i][2]}` : ''
            addTypeData(typeList, posList, DataType.SpecialLock, SpecialLockType.OpenSpaceObstacle, cfg.openSpaceObstacle[i], 0, [1, 1], undefined, subTypeArg)
        }
    }
    if (cfg.openSpace.length > 0) { // 空地
        for (let i = 1; i <= cfg.openSpace[0]; i++) {
            addTypeData(typeList, posList, DataType.SpecialLock, SpecialLockType.OpenSpace, cfg.openSpace[i], 0)
        }
    }

    if (cfg.chain.length > 0) { // 锁链
        for (let i = 1; i <= cfg.chain[0]; i++) {
            const times = cfg.chain[i] && cfg.chain[i][2] ? cfg.chain[i][2] : 1;
            addTypeData(typeList, posList, DataType.NormalLock, NormalLockType.Chain, cfg.chain[i], times)
        }
    }
    if (cfg.ice.length > 0) { // 冰块
        for (let i = 1; i <= cfg.ice[0]; i++) {
            const times = cfg.ice[i] && cfg.ice[i][2] ? cfg.ice[i][2] : 1;
            addTypeData(typeList, posList, DataType.NormalLock, NormalLockType.Ice, cfg.ice[i], times)
        }
    }
    if (cfg.mailBox.length > 0) { // 信箱
        for (let i = 1; i <= cfg.mailBox[0]; i++) {
            addTypeData(typeList, posList, DataType.NormalLock, NormalLockType.MailBox, cfg.mailBox[i], 999)
        }
    }
    if (cfg.foam.length > 0) { // 泡沫
        for (let i = 1; i <= cfg.foam[0]; i++) {
            addTypeData(typeList, posList, DataType.NormalLock, NormalLockType.Foam, cfg.foam[i], 1)
        }
    }
    if (cfg.clownfish.length > 0) { // 小丑鱼
        for (let i = 1; i <= cfg.clownfish[0]; i++) {
            addTypeData(typeList, posList, DataType.NormalLock, NormalLockType.Clownfish, cfg.clownfish[i], 999)
        }
    }
    const dataTypeList = world.fetchSingle(DataTypeList).data;
    for (let { type, subType, pos, size, needBombTimes, subTypeArg } of typeList) {
        if (!pos) {
            pos = randomPos(posList, dataTypeList.rand, size);
        }
        let [x, y] = pos;
        let oldId = layerIds[x][y];
        if (oldId) throw new Error('同一个坐标配置了多个锁层物品' + [x, y] + '物品类型' + subType);
        let id = createOneData(world, pos, 1, LayerType.Lock, size, needBombTimes, type, subType, subTypeArg);
        for (let i = 0; i < size[0]; i++) {
            for (let j = 0; j < size[1]; j++) {
                layerIds[x + i][y + j] = id;
            }
        }
    }
    return layerIds;
}


// 初始化物品层
const initDataLayer = (world: World, lockLayer: number[][], cfg: MatchMapCfg, rand: Rand, w: number, h: number) => {
    const typeC = world.fetchComponent(Entity, TypeC);
    const globalState = world.fetchSingle(GlobalState).data;
    const heroSite = world.fetchSingle(HeroSite).data;
    const lockTypeC = world.fetchComponent(Entity, LockTypeC);
    let layerIds = [];
    let posList = [];
    let typeList = [];
    for (let x = 0; x < w; x++) {
        layerIds[x] = []
        for (let y = 0; y < h; y++) {
            layerIds[x][y] = 0;
            let lockId = lockLayer[x][y];
            // 不能炸 | 不能压住物品
            if (lockId && (typeC.get(lockId).cannotBomb || lockTypeC.get(lockId).cannotHoldData)) continue;
            posList.push([x, y]);
        }
    }
    if (cfg.emptyBox.length > 0) { // 空箱子
        for (let i = 1; i <= cfg.emptyBox[0]; i++) {
            addTypeData(typeList, posList, DataType.EmptyBox, null, cfg.emptyBox[i], 1)
        }
    }
    if (cfg.empty && cfg.empty.length > 0) { // 空位置
        for (let i = 1; i <= cfg.empty[0]; i++) {
            addTypeData(typeList, posList, null, null, cfg.empty[i], 1)
        }
    }
    if (cfg.fixPos.length > 0) { // 固定特殊物品
        for (let i = 0; i < cfg.fixPos.length; i++) {
            let [type, x, y] = cfg.fixPos[i];
            let needBombTimes = type === DataType.Medusa ? MedusaNeedBombTimes : 1;
            addTypeData(typeList, posList, type, null, [x, y], needBombTimes)
        }
    }
    if (cfg.synthesisData.length > 0) { // 合成物品
        for (let i = 0; i < cfg.synthesisData.length; i++) {
            let [type, x, y] = cfg.synthesisData[i];
            addTypeData(typeList, posList, DataType.SynthesisData, type, [x, y], 0)
        }
    }
    let fixAwardData = cfg.fixAwardData.slice();
    if (cfg.awardData.length > 0) { // 奖励物品
        for (let [type, count, award] of cfg.awardData) {
            for (let i = 0; i < count; i++) {
                let ind = fixAwardData.findIndex(v => v[0] === award); // 固定坐标
                let pos = null;
                if (ind > -1) {
                    let [award, x, y] = fixAwardData[ind];
                    fixAwardData.splice(ind, 1);
                    pos = [x, y];
                }
                addTypeData(typeList, posList, null, null, pos, 1, [1, 1], type, award)
            }
        }
    }
    if (cfg.fixEventData.length > 0) { // 固定事件物品坐标
        for (let [award, x, y] of cfg.fixEventData) {
            let ind = globalState.eventDataList.findIndex(v => v && v[1] === award);
            if (ind > -1) {
                let [type] = globalState.eventDataList[ind];
                // 英雄 
                if (type === EventDataType.Hero) {
                    award = randomHeroSite(heroSite, award)
                }
                globalState.eventDataList[ind] = null;
                addTypeData(typeList, posList, DataType.EventData, type, [x, y], 1, [1, 1], null, award)
            }
        }
    }

    for (let { type, subType, pos, size, needBombTimes, subTypeArg, hideType } of typeList) {
        if (!pos) {
            pos = randomPos(posList, rand, size);
        }
        let [x, y] = pos;
        let oldId = layerIds[x][y];
        if (oldId) throw new Error('同一个坐标配置了多个物品层物品' + [x, y] + '物品类型' + subType);
        let id = createOneData(world, pos, 1, LayerType.Data, size, needBombTimes, type, subType, subTypeArg, hideType);
        for (let i = 0; i < size[0]; i++) {
            for (let j = 0; j < size[1]; j++) {
                layerIds[x + i][y + j] = id;
            }
        }
    }
    // 剩下位置随机物品
    for (let [x, y] of posList) {
        let id = createOneData(world, [x, y], 1, LayerType.Data);
        layerIds[x][y] = id;
    }
    return layerIds;
}


// 初始化草坪层
const initLawnLayer = (world: World, lockLayer: number[][], cfg: MatchMapCfg, rand: Rand, w: number, h: number) => {
    const typeC = world.fetchComponent(Entity, TypeC);
    const lockTypeC = world.fetchComponent(Entity, LockTypeC);
    const goalC = world.fetchComponent(Entity, GoalC);
    let layerIds = [];
    let posList = [];
    for (let x = 0; x < w; x++) {
        layerIds[x] = []
        for (let y = 0; y < h; y++) {
            layerIds[x][y] = 0;
            let lockId = lockLayer[x][y];
            // 不能炸 | 不能压住物品
            if (lockId && (typeC.get(lockId).cannotBomb || lockTypeC.get(lockId).cannotHoldData)) continue;
            posList.push([x, y]);
        }
    }
    if (cfg.haystack.length > 0) { // 草垛 全固定位置
        for (let i = 0; i < cfg.haystack.length; i++) {
            let { pos, count, list } = cfg.haystack[i];
            let id = createOneData(world, pos, 1, LayerType.Lawn, [1, 1], count, DataType.Haystack);
            goalC.insert(id, new GoalC(GoalType.ShowData, null, 0, { list: list.slice(), count }));
            updatePosList(posList, layerIds, pos, [1, 1], id);
        }
    }
    if (cfg.lawn.length > 0) { // 草坪
        for (let i = 1; i <= cfg.lawn[0]; i++) {
            let pos = cfg.lawn[i];
            if (!pos) {
                pos = randomPos(posList, rand)
            }
            updatePosList(posList, layerIds, pos, [1, 1], -1);
        }
    }
    return layerIds;
}

// 替换基础物品
const replaceDataLayer = (world: World, cfg: MatchMapCfg, dataLayer: number[][]) => {
    const typeC = world.fetchComponent(Entity, TypeC);
    const globalState = world.fetchSingle(GlobalState).data;
    const stateC = world.fetchComponent(Entity, StateC);
    const bombList = world.fetchSingle(BombList).data;
    if (cfg.replaceBasePos.length > 0) {
        for (let [type, x, y] of cfg.replaceBasePos) {
            let id = dataLayer[x][y];
            if (!id || typeC.get(id).value > DataType.EmptyBox) throw new Error('只能替换基础物品，错误坐标' + [x, y])
            typeC.get(id).value = type;
        }
    }
    if (cfg.replaceSynthesisBomb.length > 0) {
        for (let [type, x, y] of cfg.replaceSynthesisBomb) {
            let id = dataLayer[x][y];
            if (!id) continue
            // 销毁原位置上的物品
            // if (id) stateC.get(id).value = DataState.Exploded;
            let newId = createOneData(world, [x, y], globalState.curArea, LayerType.Data, [1, 1], 1, DataType.SynthesisBomb, type);
            dataLayer[x][y] = newId;
            console.log('合成一级炸弹', newId, [x, y], BombType[type]);
            bombList.waitSynthesis.push(newId);
        }
    }
}

const addTypeData = (typeList: any[], posList: number[][], type: DataType, subType: NormalLockType | SpecialLockType, pos: number[], needBombTimes: number, size: [number, number] = [1, 1], hideType?: any, subTypeArg?: string) => {
    if (type) typeList.push({ type, subType, pos, size, needBombTimes, hideType, subTypeArg })
    if (pos) {
        for (let i = 0; i < size[0]; i++) {
            for (let j = 0; j < size[1]; j++) {
                let index = posList.findIndex(v => v[0] === pos[0] + i && v[1] === pos[1] + j);
                if (index > -1) posList.splice(index, 1)
            }
        }
    }
}

const updatePosList = (posList: number[][], layerIds: number[][], pos: number[], size: number[], id: number) => {
    let [x, y] = pos;
    for (let i = 0; i < size[0]; i++) {
        for (let j = 0; j < size[1]; j++) {
            layerIds[x + i][y + j] = id;
            let index = posList.findIndex(v => v[0] === x + i && v[1] === y + j);
            if (index > -1) posList.splice(index, 1)
        }
    }
}

// 随机英雄位置
const randomHeroSite = (heroSite: HeroSite, site: number) => {
    let heroList = heroSite.heroList;
    // 该位置上有英雄
    if (heroList[site - 1]) return site;

    let frontList = [];
    let backList = [];
    let maxIndex = [0, 0];
    let siteMap = heroSite.siteMap;
    for (let i = 0; i < heroList.length; i++) {
        if (!heroList[i]) continue;
        let weight = siteMap.get(i + 1) || 0;
        // 未出现过的权重更高
        weight = 100 - weight;
        if (i < heroList.length / 2) {
            frontList.push([i + 1, weight]);
            if (weight > frontList[maxIndex[0]][1]) maxIndex[0] = frontList.length - 1;
        }
        else {
            backList.push([i + 1, weight]);
            if (weight > backList[maxIndex[1]][1]) maxIndex[1] = backList.length - 1;
        }
    }
    /**
     * @note 随机都改为按顺序添加
     */
    // 前排随机
    if (site > 0 && site <= heroList.length / 2 && frontList.length > 0) {
        site = frontList[maxIndex[0]][0]
    }
    // 后排随机
    else if (site > heroList.length / 2 && backList.length > 0) {
        site = backList[maxIndex[1]][0]
    }
    // 全队列随机
    else {
        let siteList = frontList.concat(backList);
        if (siteList.length > 0) {
            let front = frontList[maxIndex[0]]?.[1] || 0;
            let back = backList[maxIndex[1]]?.[1] || 0
            if (front > back) site = frontList[maxIndex[0]][0]
            else site = backList[maxIndex[1]][0]
        }
    }
    let count = siteMap.get(site) || 0;
    siteMap.set(site, count + 1);
    console.log('randomHeroSite siteMap === ', site, siteMap);
    heroSite.siteMap = siteMap;

    return site;
}